"""
Settings module.

This module defines the settings classes.
"""

import importlib
import itertools
import json
import os
import warnings
from pathlib import Path
from typing import Callable, Iterable, List, Optional, cast

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.core.validators import MaxLengthValidator, MaxValueValidator, MinLengthValidator, MinValueValidator

from .validators import (
    DictKeysTypeValidator,
    DictValuesTypeValidator,
    FileValidator,
    TypeValidator,
    ValuesTypeValidator,
)


class Setting(object):
    """
    Base setting class.

    The setting's name and prefix are used to specify the variable name to
    look for in the project settings. The default value is returned only if
    the variable is missing and the setting is not required. If the setting
    is missing and required, trying to access it will raise an AttributeError.

    When accessing a setting's value, the value is first fetched from environment
    or the project settings, then passed to a transform function that will return
    it modified (or not). By default, no transformation is applied.

    The call_default parameter tells if we should try to call the given
    default value before returning it. This allows to give lambdas or callables
    as default values. The transform_default parameter tells if we should
    transform the default value as well through the transform method.

    If the setting value is taken from environment, decode_environ method is
    called. By default it just decodes string as JSON and throws JSONDecodeError
    if the value is not a valid JSON. Some subclasses override and extend this
    method to be able to handle even other common values.

    Class attributes:
        default_validators (list of callables): Default set of validators for the setting.
    """

    default_validators = ()  # type: Iterable[Callable]

    def __init__(
        self,
        name="",
        default=None,
        required=False,
        prefix="",
        call_default=True,
        transform_default=False,
        validators=(),
    ):
        """
        Initialization method.

        Args:
            name (str): the name of the setting.
            default (object): default value given to the setting.
            required (bool): whether the setting is required or not.
            prefix (str):
                the setting's prefix (overrides ``AppSettings.Meta`` prefix).
            call_default (bool): whether to call the default (if callable).
            transform_default (bool): whether to transform the default value.
            validators (list of callables): list of additional validators to use.
        """
        self.name = name
        self.default = default
        self.call_default = call_default
        self.transform_default = transform_default
        self.required = required
        self.prefix = prefix
        self.parent_setting = None  # type: Optional[Setting]
        self.nested_list_index = None
        self.validators = list(itertools.chain(self.default_validators, validators))

    def _reraise_if_required(self, err):
        if self.required:
            if isinstance(err, KeyError):
                msg = "%s setting is missing required item %s" % (cast(Setting, self.parent_setting).full_name, err)
            else:
                msg = "%s setting is required and %s" % (self.full_name, err)
            raise ImproperlyConfigured(msg)

    @property
    def full_name(self):
        """
        Property to return the full name of the setting.

        Returns:
            str: upper prefix + upper name.
        """
        return self.prefix.upper() + self.name.upper()

    @property
    def default_value(self):
        """
        Property to return the default value.

        If the default value is callable and call_default is True, return
        the result of default(). Else return default.

        Returns:
            object: the default value.
        """
        if callable(self.default) and self.call_default:
            return self.default()
        return self.default

    @property
    def raw_value(self):
        """
        Property to return the variable defined in ``os.environ`` or in``django.conf.settings``.

        Returns:
            object: the variable defined in ``os.environ`` or in ``django.conf.settings``.

        Raises:
            AttributeError: if the variable is missing.
            KeyError: if the item is missing from nested setting.
        """
        if isinstance(self.parent_setting, NestedDictSetting):
            return self.parent_setting.raw_value[self.full_name]
        elif isinstance(self.parent_setting, NestedListSetting):
            return self.parent_setting.raw_value[self.nested_list_index]
        elif self.full_name in os.environ:
            return self.decode_environ(os.environ[self.full_name])
        else:
            return getattr(settings, self.full_name)

    @property
    def value(self):
        """
        Property to return the transformed raw or default value.

        This property is a simple shortcut for get_value().

        Returns:
            object: the transformed raw value.

        Raises:
            ImproperlyConfigured: The setting doesn't have a valid value.
        """
        return self.get_value()

    def get_value(self):
        """
        Return the transformed raw or default value.

        If the variable is missing from the project settings, and the setting
        is required, re-raise an AttributeError. If it is not required,
        return the (optionally transformed) default value.

        Returns:
            object: the transformed raw value.

        Raises:
            ImproperlyConfigured: The setting doesn't have a valid value.
        """
        try:
            value = self.raw_value
        except (AttributeError, KeyError) as err:
            self._reraise_if_required(err)
            default_value = self.default_value
            if self.transform_default:
                return self.transform(default_value)
            return default_value
        else:
            return self.transform(value)

    def validate(self, value):
        """Run custom validation on the setting value.

        By default, no additional validation is performed.

        Raises:
            ValidationError: if the validation fails.

        """
        pass

    def run_validators(self, value):
        """Run the validators on the setting value."""
        errors = []  # type: List[str]
        for validator in self.validators:
            try:
                validator(value)
            except ValidationError as error:
                errors.extend(error.messages)
        if errors:
            raise ValidationError(errors)

    def check(self):
        """
        Validate the setting value.

        Raises:
            ImproperlyConfigured: The setting doesn't have a valid value.
        """
        try:
            value = self.raw_value
        except (AttributeError, KeyError) as err:
            self._reraise_if_required(err)
        else:
            try:
                self.validate(value)
                self.run_validators(value)
            except ValidationError as error:
                raise ImproperlyConfigured("Setting {} has an invalid value: {}".format(self.full_name, error))

    def transform(self, value):
        """
        Return a transformed value.

        By default, no transformation is done.

        Args:
            value (object):

        Returns:
            object: the transformed value.
        """
        return value

    def decode_environ(self, value):
        """
        Return a decoded value.

        By default, loads a JSON.

        Args:
            value (string):

        Returns:
            Any: the decoded value
        """
        return json.loads(value)


class BooleanSetting(Setting):
    """Boolean setting."""

    default_validators = (TypeValidator(bool),)

    def __init__(
        self,
        name="",
        default=True,
        required=False,
        prefix="",
        call_default=True,
        transform_default=False,
        validators=(),
    ):
        """
        Initialization method.

        Args:
            name (str): the name of the setting.
            default (bool): default value given to the setting.
            required (bool): whether the setting is required or not.
            prefix (str):
                the setting's prefix (overrides ``AppSettings.Meta`` prefix).
            call_default (bool): whether to call the default (if callable).
            transform_default (bool): whether to transform the default value.
            validators (list of callables): list of additional validators to use.
        """
        super().__init__(
            name=name,
            default=default,
            required=required,
            prefix=prefix,
            call_default=call_default,
            transform_default=transform_default,
            validators=validators,
        )

    def decode_environ(self, value):
        """
        Return a decoded value.

        Try to load a valid JSON. If JSONDecodeError is raised, chcecks for other common values (True, False, yes, no,
        0, 1).

        Args:
            value (string):

        Returns:
            bool:
        """
        if value.lower() in ("true", "yes", "1"):
            return True
        elif value.lower() in ("false", "no", "0"):
            return False
        else:
            raise ValueError("Invalid boolean setting %s in environ (%s)" % (self.full_name, value))


class IntegerSetting(Setting):
    """Integer setting."""

    default_validators = (TypeValidator(int),)

    def __init__(
        self,
        name="",
        default=0,
        required=False,
        prefix="",
        call_default=True,
        transform_default=False,
        validators=(),
        minimum=None,
        maximum=None,
    ):
        """
        Initialization method.

        Args:
            name (str): the name of the setting.
            default (int): default value given to the setting.
            required (bool): whether the setting is required or not.
            prefix (str):
                the setting's prefix (overrides ``AppSettings.Meta`` prefix).
            call_default (bool): whether to call the default (if callable).
            transform_default (bool): whether to transform the default value.
            validators (list of callables): list of additional validators to use.
            minimum (int): a minimum value (included).
            maximum (int): a maximum value (included).
        """
        super().__init__(
            name=name,
            default=default,
            required=required,
            prefix=prefix,
            call_default=call_default,
            transform_default=transform_default,
            validators=validators,
        )
        if minimum is not None:
            self.validators.append(MinValueValidator(minimum))
        if maximum is not None:
            self.validators.append(MaxValueValidator(maximum))


class PositiveIntegerSetting(IntegerSetting):
    """Positive integer setting."""

    def __init__(
        self,
        name="",
        default=0,
        required=False,
        prefix="",
        call_default=True,
        transform_default=False,
        validators=(),
        maximum=None,
    ):
        """
        Initialization method.

        Args:
            name (str): the name of the setting.
            default (int): default value given to the setting.
            required (bool): whether the setting is required or not.
            prefix (str):
                the setting's prefix (overrides ``AppSettings.Meta`` prefix).
            call_default (bool): whether to call the default (if callable).
            transform_default (bool): whether to transform the default value.
            validators (list of callables): list of additional validators to use.
            maximum (int): a maximum value (included).
        """
        super().__init__(
            name=name,
            default=default,
            required=required,
            prefix=prefix,
            call_default=call_default,
            transform_default=transform_default,
            validators=validators,
            minimum=0,
            maximum=maximum,
        )


class FloatSetting(IntegerSetting):
    """Float setting."""

    default_validators = (TypeValidator(float),)

    def __init__(
        self,
        name="",
        default=0.0,
        required=False,
        prefix="",
        call_default=True,
        transform_default=False,
        validators=(),
        minimum=None,
        maximum=None,
    ):
        """
        Initialization method.

        Args:
            name (str): the name of the setting.
            default (float): default value given to the setting.
            required (bool): whether the setting is required or not.
            prefix (str):
                the setting's prefix (overrides ``AppSettings.Meta`` prefix).
            call_default (bool): whether to call the default (if callable).
            transform_default (bool): whether to transform the default value.
            validators (list of callables): list of additional validators to use.
            minimum (int): a minimum value (included).
            maximum (int): a maximum value (included).
        """
        super().__init__(
            name=name,
            default=default,
            required=required,
            prefix=prefix,
            call_default=call_default,
            transform_default=transform_default,
            validators=validators,
            minimum=minimum,
            maximum=maximum,
        )


class PositiveFloatSetting(FloatSetting):
    """Positive float setting."""

    def __init__(
        self,
        name="",
        default=0.0,
        required=False,
        prefix="",
        call_default=True,
        transform_default=False,
        validators=(),
        maximum=None,
    ):
        """
        Initialization method.

        Args:
            name (str): the name of the setting.
            default (float): default value given to the setting.
            required (bool): whether the setting is required or not.
            prefix (str):
                the setting's prefix (overrides ``AppSettings.Meta`` prefix).
            call_default (bool): whether to call the default (if callable).
            transform_default (bool): whether to transform the default value.
            validators (list of callables): list of additional validators to use.
            maximum (int): a maximum value (included).
        """
        super().__init__(
            name=name,
            default=default,
            required=required,
            prefix=prefix,
            call_default=call_default,
            transform_default=transform_default,
            validators=validators,
            minimum=0,
            maximum=maximum,
        )


# Iterable settings -----------------------------------------------------------
class IterableSetting(Setting):
    """Iterable setting."""

    def __init__(
        self,
        name="",
        default=None,
        required=False,
        prefix="",
        call_default=True,
        transform_default=False,
        validators=(),
        item_type=None,
        delimiter=":",
        min_length=None,
        max_length=None,
        empty=None,
    ):
        """
        Initialization method.

        Args:
            name (str): the name of the setting.
            default (iterable): default value given to the setting.
            required (bool): whether the setting is required or not.
            prefix (str):
                the setting's prefix (overrides ``AppSettings.Meta`` prefix).
            call_default (bool): whether to call the default (if callable).
            transform_default (bool): whether to transform the default value.
            validators (list of callables): list of additional validators to use.
            item_type (type): the type of the items inside the iterable.
            delimiter (str): value used to split a string into separated parts
            min_length (int): minimum length of the iterable (included).
            max_length (int): maximum length of the iterable (included).
            empty (bool): whether empty iterable is allowed. Deprecated in favor of min_length.
        """
        super().__init__(
            name=name,
            default=default,
            required=required,
            prefix=prefix,
            call_default=call_default,
            transform_default=transform_default,
            validators=validators,
        )
        self.item_type = item_type
        self.delimiter = delimiter
        if item_type is not None:
            self.validators.append(ValuesTypeValidator(item_type))
        if empty is not None:
            warnings.warn("Empty argument is deprecated, use min_length instead.", DeprecationWarning)
            if not empty:
                min_length = 1
        if min_length is not None:
            self.validators.append(MinLengthValidator(min_length))
        if max_length is not None:
            self.validators.append(MaxLengthValidator(max_length))

    def decode_environ(self, value):
        """
        Decode JSON value or split value by a delimiter to a list, if JSONDecodeError is raised.

        Default delimiter is a colon, can be changed via attribute ``delimiter``.

        Args:
            value (str):

        Returns:
            Iterable:
        """
        try:
            value = super().decode_environ(value)
        except json.decoder.JSONDecodeError:
            split_value = value.split(self.delimiter)
            item_convert_func = self.item_type or (lambda v: v)
            value = [item_convert_func(v) for v in split_value]
        return value


class StringSetting(Setting):
    """String setting."""

    default_validators = (TypeValidator(str),)

    def __init__(
        self,
        name="",
        default="",
        required=False,
        prefix="",
        call_default=True,
        transform_default=False,
        validators=(),
        min_length=None,
        max_length=None,
        empty=None,
    ):
        """
        Initialization method.

        Args:
            name (str): the name of the setting.
            default (str): default value given to the setting.
            required (bool): whether the setting is required or not.
            prefix (str):
                the setting's prefix (overrides ``AppSettings.Meta`` prefix).
            call_default (bool): whether to call the default (if callable).
            transform_default (bool): whether to transform the default value.
            validators (list of callables): list of additional validators to use.
            min_length (int): minimum length of the iterable (included).
            max_length (int): maximum length of the iterable (included).
            empty (bool): whether empty iterable is allowed. Deprecated in favor of min_length.
        """
        super().__init__(
            name=name,
            default=default,
            required=required,
            prefix=prefix,
            call_default=call_default,
            transform_default=transform_default,
            validators=validators,
        )
        if empty is not None:
            warnings.warn("Empty argument is deprecated, use min_length instead.", DeprecationWarning)
            if not empty:
                min_length = 1
        if min_length is not None:
            self.validators.append(MinLengthValidator(min_length))
        if max_length is not None:
            self.validators.append(MaxLengthValidator(max_length))

    def decode_environ(self, value):
        """
        Return a decoded value.

        Try to load JSON or return pure string, if JSONDecodeError is raised.

        Args:
            (string):

        Returns:
            string:
        """
        try:
            value = super().decode_environ(value)
        except json.decoder.JSONDecodeError:
            value = str(value)
        return value


class ListSetting(IterableSetting):
    """List setting."""

    default_validators = (TypeValidator(list),)

    def __init__(self, name="", default=list, **kwargs):
        """
        Initialization method.

        Args:
            name (str): the name of the setting.
            default (list): default value given to the setting.
            required (bool): whether the setting is required or not.
            prefix (str):
                the setting's prefix (overrides ``AppSettings.Meta`` prefix).
            call_default (bool): whether to call the default (if callable).
            transform_default (bool): whether to transform the default value.
            validators (list of callables): list of additional validators to use.
            item_type (type): the type of the items inside the iterable.
            min_length (int): minimum length of the iterable (included).
            max_length (int): maximum length of the iterable (included).
            empty (bool): whether empty iterable is allowed. Deprecated in favor of min_length.
        """
        super().__init__(name=name, default=default, **kwargs)

    def decode_environ(self, value):
        """
        Decode JSON value or split value by a delimiter to a list, if JSONDecodeError is raised.

        Default delimiter is a colon, can be changed via attribute ``delimiter``.

        Args:
            value (str):

        Returns:
            list:
        """
        return list(super().decode_environ(value))


class SetSetting(IterableSetting):
    """Set setting."""

    default_validators = (TypeValidator(set),)

    def __init__(self, name="", default=set, **kwargs):
        """
        Initialization method.

        Args:
            name (str): the name of the setting.
            default (set): default value given to the setting.
            required (bool): whether the setting is required or not.
            prefix (str):
                the setting's prefix (overrides ``AppSettings.Meta`` prefix).
            call_default (bool): whether to call the default (if callable).
            transform_default (bool): whether to transform the default value.
            validators (list of callables): list of additional validators to use.
            item_type (type): the type of the items inside the iterable.
            min_length (int): minimum length of the iterable (included).
            max_length (int): maximum length of the iterable (included).
            empty (bool): whether empty iterable is allowed. Deprecated in favor of min_length.
        """
        super().__init__(name=name, default=default, **kwargs)

    def decode_environ(self, value):
        """
        Decode JSON value or split value by a delimiter to a set, if JSONDecodeError is raised.

        Default delimiter is a colon, can be changed via attribute ``delimiter``.

        Args:
            value (str):

        Returns:
            set:
        """
        return set(super().decode_environ(value))


class TupleSetting(IterableSetting):
    """Tuple setting."""

    default_validators = (TypeValidator(tuple),)

    def __init__(self, name="", default=tuple, **kwargs):
        """
        Initialization method.

        Args:
            name (str): the name of the setting.
            default (tuple): default value given to the setting.
            required (bool): whether the setting is required or not.
            prefix (str):
                the setting's prefix (overrides ``AppSettings.Meta`` prefix).
            call_default (bool): whether to call the default (if callable).
            transform_default (bool): whether to transform the default value.
            validators (list of callables): list of additional validators to use.
            item_type (type): the type of the items inside the iterable.
            min_length (int): minimum length of the iterable (included).
            max_length (int): maximum length of the iterable (included).
            empty (bool): whether empty iterable is allowed. Deprecated in favor of min_length.
        """
        super().__init__(name=name, default=default, **kwargs)

    def decode_environ(self, value):
        """
        Decode JSON value or split value by a delimiter to a tuple, if JSONDecodeError is raised.

        Default delimiter is a colon, can be changed via attribute ``delimiter``.

        Args:
            value (str):

        Returns:
            tuple:
        """
        return tuple(super().decode_environ(value))


# Dict settings ---------------------------------------------------------------
class DictSetting(Setting):
    """Dict setting."""

    default_validators = (TypeValidator(dict),)

    def __init__(
        self,
        name="",
        default=dict,
        required=False,
        prefix="",
        call_default=True,
        transform_default=False,
        validators=(),
        key_type=None,
        value_type=None,
        outer_delimiter=None,
        inner_delimiter="=",
        min_length=None,
        max_length=None,
        empty=None,
    ):
        """
        Initialization method.

        Args:
            name (str): the name of the setting.
            default (dict): default value given to the setting.
            required (bool): whether the setting is required or not.
            prefix (str):
                the setting's prefix (overrides ``AppSettings.Meta`` prefix).
            call_default (bool): whether to call the default (if callable).
            transform_default (bool): whether to transform the default value.
            validators (list of callables): list of additional validators to use.
            key_type: the type of the dict keys.
            value_type (type): the type of dict values.
            outer_delimiter (str): value used to split environ string into separated parts
            inner_delimiter (str): value used to split environ string parts
            min_length (int): Noop. Deprecated.
            max_length (int): Noop. Deprecated.
            empty (bool): whether empty iterable is allowed. Deprecated in favor of MinLengthValidator.
        """
        super().__init__(
            name=name,
            default=default,
            required=required,
            prefix=prefix,
            call_default=call_default,
            transform_default=transform_default,
            validators=validators,
        )
        self.key_type = key_type
        self.value_type = value_type
        self.outer_delimiter = outer_delimiter
        self.inner_delimiter = inner_delimiter
        if key_type is not None:
            self.validators.append(DictKeysTypeValidator(key_type))
        if value_type is not None:
            self.validators.append(DictValuesTypeValidator(value_type))
        if empty is not None:
            warnings.warn("Empty argument is deprecated, use MinLengthValidator instead.", DeprecationWarning)
            self.validators.append(MinLengthValidator(1))
        if min_length is not None:
            warnings.warn("Argument min_length does nothing and is deprecated.", DeprecationWarning)
        if max_length is not None:
            warnings.warn("Argument max_length does nothing and is deprecated.", DeprecationWarning)

    def decode_environ(self, value):
        """
        Decode JSON value or split value by delimiters to a dict, if JSONDecodeError is raised.

        Default delimiter to distinguish single items is a whitespace sequence, items are then split by equal sign by
        default. Both delimiters can be changed via instance attributes ``inner_delimiter`` and ``outer_delimiter``.

        Args:
            value (str):

        Raises:
            ValueError: not enough values to unpack

        Returns:
            dict:
        """
        try:
            value = super().decode_environ(value)
        except json.decoder.JSONDecodeError:
            key_func = self.key_type or (lambda v: v)
            value_func = self.value_type or (lambda v: v)
            value = {
                key_func(k): value_func(v)
                for k, v in [value.split(self.inner_delimiter, 2) for value in value.split(self.outer_delimiter)]
            }
        return value


# Complex settings ------------------------------------------------------------
class ObjectSetting(Setting):
    """
    Object setting.

    This setting allows to return an object given its Python path (a.b.c).
    """

    default_validators = (TypeValidator(str),)

    def __init__(
        self,
        name="",
        default=None,
        required=False,
        prefix="",
        call_default=True,
        transform_default=False,
        validators=(),
        min_length=None,
        max_length=None,
        empty=None,
    ):
        """
        Initialization method.

        Args:
            name (str): the name of the setting.
            default (object): default value given to the setting.
            required (bool): whether the setting is required or not.
            prefix (str):
                the setting's prefix (overrides ``AppSettings.Meta`` prefix).
            call_default (bool): whether to call the default (if callable).
            transform_default (bool): whether to transform the default value.
            validators (list of callables): list of additional validators to use.
            min_length (int): Noop. Deprecated.
            max_length (int): Noop. Deprecated.
            empty (bool): Noop. Deprecated.
        """
        super().__init__(
            name=name,
            default=default,
            required=required,
            prefix=prefix,
            call_default=call_default,
            transform_default=transform_default,
            validators=validators,
        )
        if min_length is not None:
            warnings.warn("Argument min_length does nothing and is deprecated.", DeprecationWarning)
        if max_length is not None:
            warnings.warn("Argument max_length does nothing and is deprecated.", DeprecationWarning)
        if empty is not None:
            warnings.warn("Argument empty does nothing and is deprecated.", DeprecationWarning)

    def transform(self, path):
        """
        Transform a path into an actual Python object.

        The path can be arbitrary long. You can pass the path to a package,
        a module, a class, a function or a global variable, as deep as you
        want, as long as the deepest module is importable through
        ``importlib.import_module`` and each object is obtainable through
        the ``getattr`` method. Local objects will not work.

        Args:
            path (str): the dot-separated path of the object.

        Returns:
            object: the imported module or obtained object.
        """
        if path is None or not path:
            return None

        obj_parent_modules = path.split(".")
        objects = [obj_parent_modules.pop(-1)]

        while True:
            try:
                parent_module_path = ".".join(obj_parent_modules)
                parent_module = importlib.import_module(parent_module_path)
                break
            except ImportError:
                if len(obj_parent_modules) == 1:
                    raise ImportError("No module named '%s'" % obj_parent_modules[0])
                objects.insert(0, obj_parent_modules.pop(-1))

        current_object = parent_module
        for obj in objects:
            current_object = getattr(current_object, obj)
        return current_object

    def decode_environ(self, value):
        """
        Try to load JSON or return pure string, if JSONDecodeError is raised.

        Args:
            (string):

        Returns:
            string:
        """
        try:
            value = super().decode_environ(value)
        except json.decoder.JSONDecodeError:
            value = str(value)
        return value


# Callable path settings ------------------------------------------------------
class CallablePathSetting(ObjectSetting):
    """
    Callable path setting.

    This setting value should be string containing a dotted path to a callable.
    """

    def validate(self, value):
        """
        Check whether the value is path to a callable.

        We disregard the raw value and use transformed value instead.
        """
        super().validate(value)
        transformed_value = self.value
        if not callable(transformed_value):
            raise ValidationError("Value %(value)s is not a callable.", params={"value": transformed_value})


# Nested settings -------------------------------------------------------------
class NestedDictSetting(DictSetting):
    """
    Nested dict setting.

    Environment variables are not passed to inner settings.
    """

    def __init__(self, settings, *args, **kwargs):
        """
        Initialization method.

        Args:
            settings (dict): subsettings.
            name (str): the name of the setting.
            default (dict): default value given to the setting.
            required (bool): whether the setting is required or not.
            prefix (str):
                the setting's prefix (overrides ``AppSettings.Meta`` prefix).
            call_default (bool): whether to call the default (if callable).
            transform_default (bool): whether to transform the default value.
            validators (list of callables): list of additional validators to use.
            key_type: the type of the dict keys.
            value_type (type): the type of dict values.
            min_length (int): minimum length of the iterable (included).
            max_length (int): maximum length of the iterable (included).
            empty (bool): whether empty iterable is allowed. Deprecated in favor of min_length.
        """
        super().__init__(*args, **kwargs)
        for subname, subsetting in settings.items():
            if subsetting.name == "":
                subsetting.name = subname
            subsetting.parent_setting = self
        self.settings = settings

    def get_value(self):
        """
        Return dictionary with values of subsettings.

        Returns:
            dict: values of subsettings.
        """
        try:
            self.raw_value
        except (AttributeError, KeyError) as err:
            self._reraise_if_required(err)
            default_value = self.default_value
            if self.transform_default:
                return self.transform(default_value)
            return default_value
        else:
            # If setting is defined, load values of all subsettings.
            value = {}
            for key, subsetting in self.settings.items():
                value[key] = subsetting.get_value()
            return value

    def check(self):
        """
        Validate the setting value.

        Raises:
            AttributeError: if the setting is missing and required.
            ValueError: (or other Exception) if the raw value is invalid.
        """
        super().check()
        errors = []  # type: List[str]
        try:
            raw_value = self.raw_value
        except AttributeError:
            # If not required and not passed
            pass
        else:
            if raw_value is not None:
                for subsetting in self.settings.values():
                    try:
                        subsetting.check()
                    except ValidationError as error:
                        errors.extend(error.messages)
                if errors:
                    raise ValidationError(errors)


class NestedSetting(NestedDictSetting):
    """Deprecated alias for NestedDictSetting."""

    def __init__(self, *args, **kwargs):
        """
        Initialization method.

        Args:
            settings (dict): subsettings.
            name (str): the name of the setting.
            default (dict): default value given to the setting.
            required (bool): whether the setting is required or not.
            prefix (str):
                the setting's prefix (overrides ``AppSettings.Meta`` prefix).
            call_default (bool): whether to call the default (if callable).
            transform_default (bool): whether to transform the default value.
            validators (list of callables): list of additional validators to use.
            key_type: the type of the dict keys.
            value_type (type): the type of dict values.
            min_length (int): minimum length of the iterable (included).
            max_length (int): maximum length of the iterable (included).
            empty (bool): whether empty iterable is allowed. Deprecated in favor of min_length.
        """
        super().__init__(*args, **kwargs)
        warnings.warn("NestedSetting is deprecated in favor of NestedDictSetting.", DeprecationWarning)


class NestedListSetting(IterableSetting):
    """
    Nested list setting.

    Environment variables are not passed to inner settings.
    """

    def __init__(self, inner_setting, *args, **kwargs):
        """
        Initialization method.

        Args:
            inner_setting (Setting): setting that should be applied to list items.
                NestedDictSetting is not supported at the moment.
            name (str): the name of the setting.
            default (dict): default value given to the setting.
            required (bool): whether the setting is required or not.
            prefix (str):
                the setting's prefix (overrides ``AppSettings.Meta`` prefix).
            call_default (bool): whether to call the default (if callable).
            transform_default (bool): whether to transform the default value.
            validators (list of callables): list of additional validators to use.
            key_type: the type of the dict keys.
            value_type (type): the type of dict values.
            min_length (int): minimum length of the iterable (included).
            max_length (int): maximum length of the iterable (included).
            empty (bool): whether empty iterable is allowed. Deprecated in favor of min_length.
        """
        super().__init__(*args, **kwargs)
        if not inner_setting.name:
            inner_setting.name = self.name
        inner_setting.parent_setting = self
        self.inner_setting = inner_setting

    def get_value(self):
        try:
            value = self.raw_value
        except (AttributeError, KeyError) as err:
            self._reraise_if_required(err)
            default_value = self.default_value
            if self.transform_default:
                return self.transform(default_value)
            return default_value
        else:
            return_value = []
            for index, item in enumerate(value):
                self.inner_setting.nested_list_index = index
                return_value.append(self.inner_setting.get_value())
            return tuple(return_value)

    def check(self):
        """Check the nested list setting itself and all its items."""
        super().check()
        try:
            value = self.raw_value
        except (AttributeError, KeyError) as err:
            self._reraise_if_required(err)
        else:
            errors = []  # type: List[str]
            for index, item in enumerate(value):
                try:
                    self.inner_setting.nested_list_index = index
                    self.inner_setting.check()
                except ValidationError as error:
                    errors.extend(error.messages)
            if errors:
                raise ValidationError(errors)

    def transform(self, value):
        """Transform each item in the list."""
        return tuple(self.inner_setting.transform(item) for item in value)


class FileSetting(Setting):
    """
    File setting.

    Value of this setting is a pathlib.Path instance.
    """

    def __init__(
        self,
        name="",
        default=None,
        required=False,
        prefix="",
        call_default=True,
        transform_default=False,
        validators=(),
        mode=None,
    ):
        """
        Initialization method.

        Note that attribute ``mode`` is used in ``FileValidator``. For more information check its documentation.

        Args:
            name (str): the name of the setting.
            default (object): default value given to the setting.
            required (bool): whether the setting is required or not.
            prefix (str):
                the setting's prefix (overrides ``AppSettings.Meta`` prefix).
            call_default (bool): whether to call the default (if callable).
            transform_default (bool): whether to transform the default value.
            validators (list of callables): list of additional validators to use.
            mode (optional int): Required permission for the file. None means no validation (f.e. file does not exist
                yet), other values might consist of inclusive OR between F_OK, R_OK, W_OK and X_OK constants. When
                checking read, write or execute permission, file existence (F_OK) is implicitly required. For more
                information check ``os.access`` documentation (https://docs.python.org/3/library/os.html#os.access).
        """
        super().__init__(
            name=name,
            default=default,
            required=required,
            prefix=prefix,
            call_default=call_default,
            transform_default=transform_default,
            validators=validators,
        )
        if mode is not None:
            self.validators.append(FileValidator(mode))

    def transform(self, value):
        """Transform value to Path instance."""
        return Path(super().transform(value))



"""Main test script."""
import os
import tempfile
from pathlib import Path
from typing import Dict, cast
from unittest import mock

import pytest
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.test import SimpleTestCase, override_settings

import appsettings


def imported_object():
    return "tests.test_appsettings.ImportedClass._imported_object2"


class ImportedClass:
    """Mixin for tests."""

    @staticmethod
    def _imported_object2():
        return "nothing"


class SettingTestCase(SimpleTestCase):
    NOT_A_CALLABLE = {}  # type: dict

    def setUp(self):
        self.message_required = "%s setting is required and"
        self.message_missing_item = "%s setting is missing required item"
        self.message_no_attr = "has no attribute '%s'"

    def test_setting(self):
        setting = appsettings.Setting(name="simple")
        assert setting.name == "simple"
        assert setting.full_name == "SIMPLE"
        assert setting.default_value is None
        assert setting.value is None
        assert setting.get_value() is None
        assert setting.validators == []
        setting.check()
        with pytest.raises(AttributeError, match="SIMPLE"):
            assert setting.raw_value

    def test_setting_name(self):
        setting = appsettings.Setting(name="Name", prefix="Prefix_")
        assert setting.name == "Name"
        assert setting.prefix == "Prefix_"
        assert setting.full_name == "PREFIX_NAME"

    def test_setting_default_callable(self):
        setting = appsettings.Setting(default=lambda: 1, call_default=True)
        assert setting.value == 1
        setting.call_default = False
        assert callable(setting.value)
        assert setting.value() == 1

    def test_setting_default_dont_raise_exception(self):
        setting = appsettings.IntegerSetting(name="setting", default="hello")
        assert setting.value == "hello"
        with override_settings(SETTING=0):
            assert setting.value == 0

    def test_setting_required(self):
        setting = appsettings.Setting(name="setting", prefix="custom_", required=True, default=True)
        with pytest.raises(ImproperlyConfigured, match=self.message_required % setting.full_name):
            assert setting.value
        assert setting.default_value

        setting.parent_setting = appsettings.NestedDictSetting(settings={}, name="parent_setting")
        with override_settings(PARENT_SETTING={}):
            with pytest.raises(
                ImproperlyConfigured, match=self.message_missing_item % setting.parent_setting.full_name
            ):
                assert setting.value

    def test_setting_transform(self):
        class Setting(appsettings.Setting):
            def transform(self, value):
                if value is None:
                    return "TRANSFORMED"
                return value

        setting = Setting(name="transform")
        with override_settings(TRANSFORM=None):
            assert setting.value == "TRANSFORMED"
        with override_settings(TRANSFORM=1024):
            assert setting.value == 1024

    def test_setting_validators(self):
        # Test default and custom validators are correctly chained.
        class TestSetting(appsettings.Setting):
            default_validators = (mock.sentinel.validator,)

        setting = TestSetting(name="INQUISITOR", validators=(mock.sentinel.custom_validator,))
        assert setting.validators == [mock.sentinel.validator, mock.sentinel.custom_validator]

    def test_setting_validators_pass(self):
        validator = mock.Mock()
        setting = appsettings.Setting(name="INQUISITOR", validators=(validator,))

        with self.settings(INQUISITOR=mock.sentinel.lister):
            setting.check()

        assert validator.mock_calls == [mock.call(mock.sentinel.lister)]

    def test_setting_validators_fail(self):
        validator = mock.Mock(side_effect=ValidationError("You're not worthy!"))
        setting = appsettings.Setting(name="INQUISITOR", validators=(validator,))

        with self.settings(INQUISITOR=mock.sentinel.lister):
            with pytest.raises(
                ImproperlyConfigured, match="Setting INQUISITOR has an invalid value:.*You're not worthy!"
            ):
                setting.check()

        assert validator.mock_calls == [mock.call(mock.sentinel.lister)]

    def test_setting_custom_validate(self):
        # Test custom validate method
        class TestSetting(appsettings.Setting):
            def validate(self, value):
                raise ValidationError("You're not worthy!")

        setting = TestSetting(name="INQUISITOR")

        with self.settings(INQUISITOR=mock.sentinel.lister):
            with pytest.raises(
                ImproperlyConfigured, match="Setting INQUISITOR has an invalid value:.*You're not worthy!"
            ):
                setting.check()

    def test_setting_raw_value(self):
        setting = appsettings.Setting(name="setting")
        setting.check()
        with pytest.raises(AttributeError):
            setting.raw_value
        with override_settings(SETTING="value"):
            setting.check()
            assert setting.raw_value == "value"

        setting.parent_setting = appsettings.NestedDictSetting(settings={}, name="parent_setting")
        with override_settings(PARENT_SETTING={}):
            with pytest.raises(KeyError):
                setting.raw_value
        with override_settings(PARENT_SETTING={"SETTING": "value"}):
            setting.check()
            assert setting.raw_value == "value"

        setting.parent_setting = appsettings.NestedListSetting(inner_setting=setting, name="parent_setting")
        setting.nested_list_index = 0
        with override_settings(PARENT_SETTING=[]):
            with pytest.raises(IndexError):
                setting.raw_value
        with override_settings(PARENT_SETTING=["value"]):
            setting.check()
            assert setting.raw_value == "value"

    @mock.patch.dict(os.environ, {"PREFERENCE_SETTING": '"__ENV__"'})
    def test_preference_of_environ_values(self):
        setting = appsettings.Setting(name="preference_setting")
        with override_settings(PREFERENCE_SETTING="__OVER__"):
            setting.check()
            assert setting.value == "__ENV__"

    @mock.patch.dict(os.environ, {"SETTING": '{"key": ["v", "a", "l"]}'})
    def test_json_from_environ_value(self):
        setting = appsettings.Setting(name="setting")
        setting.check()
        assert setting.value == {"key": ["v", "a", "l"]}


class BooleanSettingTestCase(SimpleTestCase):
    """BooleanSetting tests."""

    def test_boolean_setting(self):
        setting = appsettings.BooleanSetting()
        assert setting.value is True

    @mock.patch.dict(os.environ, {"SETTING": "true"})
    def test_json_boolean_setting_from_environ_true_value(self):
        setting = appsettings.BooleanSetting(name="setting")
        setting.check()
        assert setting.value is True

    @mock.patch.dict(os.environ, {"BOOL_LOWER": "true", "BOOL_UPPER": "TRUE", "BOOL_NUM": "1", "BOOL_WORD": "yes"})
    def test_string_boolean_setting_from_environ_true_value(self):
        bool_lower = appsettings.BooleanSetting(name="bool_lower")
        bool_lower.check()
        assert bool_lower.value is True

        bool_upper = appsettings.BooleanSetting(name="bool_upper")
        bool_upper.check()
        assert bool_upper.value is True

        bool_num = appsettings.BooleanSetting(name="bool_num")
        bool_num.check()
        assert bool_num.value is True

        bool_word = appsettings.BooleanSetting(name="bool_word")
        bool_word.check()
        assert bool_word.value is True

    @mock.patch.dict(os.environ, {"BOOL_LOWER": "false", "BOOL_UPPER": "FALSE", "BOOL_NUM": "0", "BOOL_WORD": "no"})
    def test_string_boolean_setting_from_environ_false_value(self):
        bool_lower = appsettings.BooleanSetting(name="bool_lower")
        bool_lower.check()
        assert bool_lower.value is False

        bool_upper = appsettings.BooleanSetting(name="bool_upper")
        bool_upper.check()
        assert bool_upper.value is False

        bool_num = appsettings.BooleanSetting(name="bool_num")
        bool_num.check()
        assert bool_num.value is False

        bool_word = appsettings.BooleanSetting(name="bool_word")
        bool_word.check()
        assert bool_word.value is False

    @mock.patch.dict(os.environ, {"BOOL_SETTING": "invalid"})
    def test_string_boolean_setting_from_environ_invalid_value(self):
        bool_setting = appsettings.BooleanSetting(name="bool_setting")
        with pytest.raises(ValueError, match="Invalid boolean setting BOOL_SETTING"):
            bool_setting.check()


class IntegerSettingTestCase(SimpleTestCase):
    """IntegerSetting tests."""

    def test_integer_setting(self):
        setting = appsettings.IntegerSetting()
        assert setting.value == 0

    @mock.patch.dict(os.environ, {"SETTING": "123"})
    def test_integer_setting_from_environ_value(self):
        setting = appsettings.IntegerSetting(name="setting")
        setting.check()
        assert setting.value == 123
        assert type(setting.value) is int


class PositiveIntegerSettingTestCase(SimpleTestCase):
    """PositiveIntegerSetting tests."""

    def test_positive_integer_setting(self):
        setting = appsettings.PositiveIntegerSetting()
        assert setting.value == 0


class FloatSettingTestCase(SimpleTestCase):
    """FloatSetting tests."""

    def test_float_setting(self):
        setting = appsettings.FloatSetting()
        assert setting.value == 0.0

    @mock.patch.dict(os.environ, {"SETTING": "123.456"})
    def test_float_setting_from_environ_value(self):
        setting = appsettings.FloatSetting(name="setting")
        setting.check()
        assert setting.value == 123.456
        assert type(setting.value) is float


class PositiveFloatSettingTestCase(SimpleTestCase):
    """PositiveFloatSetting tests."""

    def test_positive_float_setting(self):
        setting = appsettings.PositiveFloatSetting()
        assert setting.value == 0.0


class IterableSettingTestCase(SimpleTestCase):
    """IterableSetting tests."""

    def test_iterable_setting(self):
        setting = appsettings.IterableSetting()
        assert setting.value is None

    @mock.patch.dict(os.environ, {"SETTING": "[1, 2, 3]"})
    def test_iterable_setting_from_environ_json_value(self):
        setting = appsettings.IterableSetting(name="setting")
        setting.check()
        assert setting.value == [1, 2, 3]

    @mock.patch.dict(os.environ, {"SETTING": "1:2:3"})
    def test_iterable_setting_from_environ_delimiter_value(self):
        setting = appsettings.IterableSetting(name="setting")
        setting.check()
        assert setting.value == ["1", "2", "3"]

    @mock.patch.dict(os.environ, {"SETTING": "1-2-3"})
    def test_iterable_setting_from_environ_delimiter_value_with_item_type(self):
        setting = appsettings.IterableSetting(name="setting", item_type=int, delimiter="-")
        setting.check()
        assert setting.value == [1, 2, 3]


class StringSettingTestCase(SimpleTestCase):
    """StringSetting tests."""

    def test_string_setting(self):
        setting = appsettings.StringSetting()
        assert setting.value == ""

    @mock.patch.dict(os.environ, {"SETTING": '"json-string"'})
    def test_string_setting_from_environ_json_value(self):
        setting = appsettings.StringSetting(name="setting")
        setting.check()
        assert setting.value == "json-string"

    @mock.patch.dict(os.environ, {"SETTING": "pure-string"})
    def test_string_setting_from_environ_pure_value(self):
        setting = appsettings.StringSetting(name="setting")
        setting.check()
        assert setting.value == "pure-string"


class ListSettingTestCase(SimpleTestCase):
    """ListSetting tests."""

    def test_list_setting(self):
        setting = appsettings.ListSetting()
        assert setting.value == []


class SetSettingTestCase(SimpleTestCase):
    """SetSetting tests."""

    def test_set_setting(self):
        setting = appsettings.SetSetting()
        assert setting.value == set()

    @mock.patch.dict(os.environ, {"SETTING": "a:b:b:b:c"})
    def test_set_setting_from_environ_value(self):
        setting = appsettings.SetSetting(name="setting")
        setting.check()
        assert setting.value == {"a", "b", "c"}


class TupleSettingTestCase(SimpleTestCase):
    """TupleSetting tests."""

    def test_tuple_setting(self):
        setting = appsettings.TupleSetting()
        assert setting.value == ()

    @mock.patch.dict(os.environ, {"SETTING": "a:b:c"})
    def test_tuple_setting_from_environ_value(self):
        setting = appsettings.TupleSetting(name="setting")
        setting.check()
        assert setting.value == ("a", "b", "c")


class DictSettingTestCase(SimpleTestCase):
    """DictSetting tests."""

    def test_dict_setting(self):
        setting = appsettings.DictSetting()
        assert setting.value == {}

    @mock.patch.dict(os.environ, {"SETTING": '{"a": "A", "b": "B"}'})
    def test_dict_setting_from_environ_json_value(self):
        setting = appsettings.DictSetting(name="setting")
        setting.check()
        assert setting.value == {"a": "A", "b": "B"}

    @mock.patch.dict(os.environ, {"SETTING": "a=A b=B"})
    def test_dict_setting_from_environ_delimiter_value(self):
        setting = appsettings.DictSetting(name="setting")
        setting.check()
        assert setting.value == {"a": "A", "b": "B"}

    @mock.patch.dict(os.environ, {"SETTING": "a:1--b:2"})
    def test_dict_setting_from_environ_delimiter_value_with_types(self):
        setting = appsettings.DictSetting(
            name="setting", outer_delimiter="--", inner_delimiter=":", key_type=str, value_type=int
        )
        setting.check()
        assert setting.value == {"a": 1, "b": 2}


class ObjectSettingTestCase(SimpleTestCase):
    """ObjectSetting tests."""

    def test_object_setting(self):
        setting = appsettings.ObjectSetting(name="object")
        setting.check()
        assert setting.value is None
        with override_settings(OBJECT="tests.test_appsettings.imported_object"):
            setting.check()
            assert setting.value is imported_object
        setting.default = imported_object
        setting.call_default = True
        assert setting.value == "tests.test_appsettings.ImportedClass._imported_object2"
        setting.transform_default = True
        assert setting.value is ImportedClass._imported_object2
        with override_settings(OBJECT="this_package.does_not_exist"):
            with pytest.raises(ImportError):
                assert setting.value
        with override_settings(OBJECT="tests.test_appsettings.SettingTestCase.this_function_does_not_exist"):
            with pytest.raises(AttributeError):
                assert setting.value
        with override_settings(OBJECT=""):
            assert setting.value is None
        with override_settings(OBJECT=None):
            assert setting.value is None

    @mock.patch.dict(os.environ, {"SETTING": "tests.test_appsettings.imported_object"})
    def test_object_setting_from_environ_value(self):
        setting = appsettings.ObjectSetting(name="setting")
        setting.check()
        assert setting.value is imported_object


class CallablePathSettingTestCase(SimpleTestCase):
    """CallablePathSetting tests."""

    def test_callable_path_setting(self):
        setting = appsettings.CallablePathSetting(name="callable_path")
        setting.check()
        assert setting.value is None
        with override_settings(CALLABLE_PATH="tests.test_appsettings.imported_object"):
            setting.check()
            assert setting.value is imported_object
        with override_settings(CALLABLE_PATH="tests.test_appsettings.SettingTestCase.NOT_A_CALLABLE"):
            with pytest.raises(ImproperlyConfigured):
                setting.check()
        with override_settings(CALLABLE_PATH=None):
            with pytest.raises(ImproperlyConfigured):
                setting.check()


class NestedDictSettingTestCase(SimpleTestCase):
    """NestedDictSetting tests."""

    def test_nested_setting(self):
        setting = appsettings.NestedDictSetting(settings=dict())
        assert setting.value == {}
        setting.transform_default = True
        assert setting.value == {}

        setting = appsettings.NestedDictSetting(
            name="setting",
            default={},
            settings=dict(
                bool1=appsettings.BooleanSetting(default=False),
                bool2=appsettings.BooleanSetting(name="bool3", default=True),
            ),
        )
        assert setting.value == {}

        with override_settings(SETTING={"BOOL3": False}):
            assert setting.value == {"bool1": False, "bool2": False}

    def test_nested_dict_setting_not_required_anything(self):
        outer_setting = appsettings.NestedDictSetting(
            name="outer_setting", settings=dict(inner_setting=appsettings.StringSetting(default="Default"))
        )

        # Not passed anything
        outer_setting.check()
        assert len(outer_setting.value.items()) == 0
        assert outer_setting.value.get("inner_setting") is None

        # Pass outer setting
        with override_settings(OUTER_SETTING={"INNER_FAKE_SETTING": "Fake setting value"}):
            outer_setting.check()
            assert len(outer_setting.value.items()) == 1
            assert outer_setting.value.get("inner_setting") == "Default"

        # Pass inner setting as well
        with override_settings(OUTER_SETTING={"INNER_SETTING": "Value"}):
            outer_setting.check()
            assert len(outer_setting.value.items()) == 1
            assert outer_setting.value.get("inner_setting") == "Value"

    def test_nested_dict_setting_required_outer_setting(self):
        outer_setting = appsettings.NestedDictSetting(
            name="outer_setting",
            required=True,
            settings=dict(inner_setting=appsettings.StringSetting(default="Default")),
        )

        # Not passed anything
        with pytest.raises(ImproperlyConfigured):
            outer_setting.check()

        # Pass outer setting
        with override_settings(OUTER_SETTING={"INNER_FAKE_SETTING": "Fake setting value"}):
            outer_setting.check()
            assert len(outer_setting.value.items()) == 1
            assert outer_setting.value.get("inner_setting") == "Default"

        # Pass inner setting as well
        with override_settings(OUTER_SETTING={"INNER_SETTING": "Value"}):
            outer_setting.check()
            assert len(outer_setting.value.items()) == 1
            assert outer_setting.value.get("inner_setting") == "Value"

    def test_nested_dict_setting_required_inner_setting(self):
        outer_setting = appsettings.NestedDictSetting(
            name="outer_setting", settings=dict(inner_setting=appsettings.StringSetting(required=True))
        )

        # Not passed anything
        outer_setting.check()
        assert len(outer_setting.value.items()) == 0
        assert outer_setting.value.get("inner_setting") is None

        # Pass outer setting
        with override_settings(OUTER_SETTING={"INNER_FAKE_SETTING": "Fake setting value"}):
            with pytest.raises(ImproperlyConfigured):
                outer_setting.check()

        # Pass inner setting as well
        with override_settings(OUTER_SETTING={"INNER_SETTING": "Value"}):
            outer_setting.check()
            assert len(outer_setting.value.items()) == 1
            assert outer_setting.value.get("inner_setting") == "Value"

    def test_nested_dict_setting_required_both_inner_and_outer_setting(self):
        outer_setting = appsettings.NestedDictSetting(
            name="outer_setting", required=True, settings=dict(inner_setting=appsettings.StringSetting(required=True))
        )

        # Not passed anything
        with pytest.raises(ImproperlyConfigured):
            outer_setting.check()

        # Pass outer setting
        with override_settings(OUTER_SETTING={"INNER_FAKE_SETTING": "Fake setting value"}):
            with pytest.raises(ImproperlyConfigured):
                outer_setting.check()

        # Pass inner setting as well
        with override_settings(OUTER_SETTING={"INNER_SETTING": "Value"}):
            outer_setting.check()
            assert len(outer_setting.value.items()) == 1
            assert outer_setting.value.get("inner_setting") == "Value"

    @mock.patch.dict(os.environ, {"SETTING": '{"A": "A", "B": "B"}'})
    def test_nested_dict_setting_from_environ_value(self):
        setting = appsettings.NestedDictSetting(
            settings=dict(a=appsettings.Setting(), b=appsettings.Setting(),), name="setting"
        )
        setting.check()
        assert setting.value == {"a": "A", "b": "B"}

    @mock.patch.dict(os.environ, {"SETTING": '{"A": "A", "B": "B"}', "A": "Fake A", "B": "Fake B"})
    def test_parent_setting_precedence_over_environ_value(self):
        setting = appsettings.NestedDictSetting(
            settings=dict(a=appsettings.StringSetting(), b=appsettings.StringSetting(),), name="setting"
        )
        setting.check()
        assert setting.value["a"] == "A"
        assert setting.value["b"] == "B"


class NestedListSettingTestCase(SimpleTestCase):
    """NestedListSetting tests."""

    def test_nested_list_setting(self):
        setting = appsettings.NestedListSetting(name="setting", default=[], inner_setting=appsettings.IntegerSetting())
        setting.check()
        assert setting.value == []

        with override_settings(SETTING=[0, 1, 2]):
            setting.check()
            assert setting.value == (0, 1, 2)
        with override_settings(SETTING=[0, "1", 2]):
            with pytest.raises(ImproperlyConfigured):
                setting.check()

        setting = appsettings.NestedListSetting(
            name="setting",
            default=["tests.test_appsettings.imported_object"],
            transform_default=True,
            inner_setting=appsettings.ObjectSetting(),
        )
        setting.check()
        assert setting.value == (imported_object,)
        with override_settings(
            SETTING=[
                "tests.test_appsettings.imported_object",
                "tests.test_appsettings.ImportedClass._imported_object2",
            ]
        ):
            setting.check()
            assert setting.value == (imported_object, ImportedClass._imported_object2)

    def test_nested_nested_list_setting(self):
        setting = appsettings.NestedListSetting(
            name="setting",
            default=[],
            inner_setting=appsettings.NestedListSetting(
                name="inner", default=[], inner_setting=appsettings.IntegerSetting()
            ),
        )
        setting.check()
        assert setting.value == []
        assert setting.inner_setting.name == "inner"
        with override_settings(SETTING=([1, 2, 3], [4, 5])):
            setting.check()
            assert setting.value == ((1, 2, 3), (4, 5))
        with override_settings(SETTING=[[1, 2, 3], ["x", 5]]):
            with pytest.raises(ImproperlyConfigured):
                setting.check()

        setting = appsettings.NestedListSetting(
            name="setting",
            inner_setting=appsettings.NestedListSetting(
                inner_setting=appsettings.NestedListSetting(inner_setting=appsettings.ObjectSetting())
            ),
        )
        assert setting.inner_setting.name == "setting"
        with override_settings(
            SETTING=[
                (
                    ["tests.test_appsettings.imported_object"],
                    ["tests.test_appsettings.ImportedClass._imported_object2"],
                )
            ]
        ):
            setting.check()
            assert setting.value == (((imported_object,), (ImportedClass._imported_object2,)),)
        with override_settings(
            SETTING=[[["tests.test_appsettings.imported_object"], ["tests.test_appsettings.object_does_not_exist"]]]
        ):
            with pytest.raises(AttributeError):
                assert setting.value

    def test_nested_list_in_nested_dict_setting(self):
        setting = appsettings.NestedDictSetting(
            name="setting",
            default={},
            settings=dict(
                select=appsettings.NestedListSetting(
                    name="pick", default=[1], inner_setting=appsettings.IntegerSetting()
                )
            ),
        )
        setting.check()
        assert setting.value == {}
        with override_settings(SETTING={}):
            setting.check()
            assert setting.value == {"select": [1]}
        with override_settings(SETTING={"PICK": [2]}):
            setting.check()
            assert setting.value == {"select": (2,)}
        with override_settings(SETTING={"PICK": ["xyz"]}):
            with pytest.raises(ImproperlyConfigured):
                setting.check()


class FileSettingTestCase(SimpleTestCase):
    def test_file_setting_string(self):
        setting = appsettings.FileSetting(name="file")
        with tempfile.NamedTemporaryFile() as fd:
            with override_settings(FILE=fd.name):
                setting.check()
                assert isinstance(setting.value, Path)
                assert setting.value.samefile(fd.name)

    def test_file_setting_path(self):
        setting = appsettings.FileSetting(name="file")
        with tempfile.NamedTemporaryFile() as fd:
            with override_settings(FILE=Path(fd.name)):
                setting.check()
                assert isinstance(setting.value, Path)
                assert setting.value.samefile(fd.name)

    def test_file_setting_with_directory(self):
        setting = appsettings.FileSetting(name="file")
        with tempfile.TemporaryDirectory() as td_name:
            with override_settings(FILE=Path(td_name)):
                setting.check()
                assert isinstance(setting.value, Path)
                assert setting.value.samefile(td_name)


class AppSettingsTestCase(SimpleTestCase):
    def test_instantiation(self):
        class AppConf(appsettings.AppSettings):
            setting = appsettings.Setting()

        appconf = AppConf()
        assert appconf
        assert appconf.setting == AppConf.setting.get_value()
        assert AppConf.setting is AppConf.settings["setting"]
        assert AppConf.settings is AppConf._meta.settings
        with pytest.raises(AttributeError):
            assert not AppConf.not_a_setting

        with pytest.raises(RuntimeError):
            assert not appsettings.AppSettings()

    def test_populating_name(self):
        class AppConf(appsettings.AppSettings):
            without_name = appsettings.Setting()
            with_name = appsettings.Setting(name="got_a_name")

        assert "without_name" in AppConf.settings
        assert "with_name" in AppConf.settings
        assert AppConf.settings["without_name"].name == "without_name"
        assert AppConf.settings["with_name"].name == "got_a_name"

    def test_populating_prefix(self):
        class NoMetaAppConf(appsettings.AppSettings):
            without_prefix = appsettings.Setting()
            with_prefix = appsettings.Setting(prefix="got_a_prefix")

        assert "without_prefix" in NoMetaAppConf.settings
        assert "with_prefix" in NoMetaAppConf.settings
        assert NoMetaAppConf.settings["without_prefix"].prefix == ""
        assert NoMetaAppConf.settings["with_prefix"].prefix == "got_a_prefix"

        class MetaAppConf(appsettings.AppSettings):
            without_prefix = appsettings.Setting()
            with_prefix = appsettings.Setting(prefix="got_a_prefix")

            class Meta:
                setting_prefix = "meta_prefix_"

        assert "without_prefix" in MetaAppConf.settings
        assert "with_prefix" in MetaAppConf.settings
        assert MetaAppConf.settings["without_prefix"].prefix == "meta_prefix_"
        assert MetaAppConf.settings["with_prefix"].prefix == "got_a_prefix"

    def test_full_name(self):
        class AppConf(appsettings.AppSettings):
            setting = appsettings.Setting(name="name")

            class Meta:
                setting_prefix = "prefix_"

        assert AppConf.settings["setting"].full_name == "PREFIX_NAME"

    def test_caching(self):
        class AppConf(appsettings.AppSettings):
            my_int = appsettings.IntegerSetting()

        appconf = AppConf()
        assert "my_int" not in appconf._cache
        assert appconf.my_int == 0
        assert "my_int" in appconf._cache
        assert appconf._cache["my_int"] == 0
        assert appconf.my_int == 0
        appconf.invalidate_cache()
        assert "my_int" not in appconf._cache
        with pytest.raises(AttributeError):
            assert not appconf.not_a_setting

    def test_invalidate_on_signal(self):
        class AppConf(appsettings.AppSettings):
            my_int = appsettings.IntegerSetting()

        appconf = AppConf()
        assert "my_int" not in appconf._cache
        assert appconf.my_int == 0
        assert "my_int" in appconf._cache
        assert appconf._cache["my_int"] == 0

        with override_settings(MY_INT=1):
            assert "my_int" not in appconf._cache
            assert appconf.my_int == 1
            assert "my_int" in appconf._cache
            assert appconf._cache["my_int"] == 1

        assert "my_int" not in appconf._cache
        assert appconf.my_int == 0

    @mock.patch.dict(os.environ, {"ONE": "Env_1", "TWO": "Env_2", "THREE": "Env_3"})
    def test_environ_values_invalidation(self):
        class AppConf(appsettings.AppSettings):
            one = appsettings.StringSetting()
            two = appsettings.StringSetting()
            three = appsettings.StringSetting()
            four = appsettings.StringSetting(default="Def_4")

        appconf = AppConf()
        with override_settings(ONE="One", TWO="Two"):
            assert "ONE" not in os.environ
            assert "__DAP_ONE" in os.environ
            assert appconf.one == "One"

            assert "TWO" not in os.environ
            assert "__DAP_TWO" in os.environ
            assert appconf.two == "Two"

            assert "THREE" in os.environ
            assert "__DAP_THREE" not in os.environ
            assert appconf.three == "Env_3"

            assert "FOUR" not in os.environ
            assert "__DAP_FOUR" not in os.environ
            assert appconf.four == "Def_4"

        assert "ONE" in os.environ
        assert "__DAP_ONE" not in os.environ
        assert appconf.one == "Env_1"

        assert "TWO" in os.environ
        assert "__DAP_TWO" not in os.environ
        assert appconf.two == "Env_2"

        assert "THREE" in os.environ
        assert "__DAP_THREE" not in os.environ
        assert appconf.three == "Env_3"

        assert "FOUR" not in os.environ
        assert "__DAP_FOUR" not in os.environ
        assert appconf.four == "Def_4"

    @mock.patch.dict(os.environ, {"SETTING": "ONE=Env_1 TWO=Env_2"})
    def test_environ_nested_setting_invalidation(self):
        class AppConf(appsettings.AppSettings):
            setting = cast(
                Dict[str, str],
                appsettings.NestedDictSetting(
                    settings=dict(
                        one=appsettings.StringSetting(required=True), two=appsettings.StringSetting(default="Def_2"),
                    ),
                    required=True,
                ),
            )

        appconf = AppConf()
        assert appconf.setting["one"] == "Env_1"
        assert appconf.setting["two"] == "Env_2"
        with override_settings(SETTING={"ONE": "One"}):
            assert "SETTING" not in os.environ
            assert "__DAP_SETTING" in os.environ
            assert appconf.setting["one"] == "One"
            assert appconf.setting["two"] == "Def_2"

        assert appconf.setting["one"] == "Env_1"
        assert appconf.setting["two"] == "Env_2"

    def test_check(self):
        assert appsettings.AppSettings.check() is None

        class AppConf(appsettings.AppSettings):
            setting = appsettings.Setting()

        assert AppConf.check() is None

        AppConf.setting.required = True

        with pytest.raises(ImproperlyConfigured):
            assert not AppConf.check()



"""Test settings validators."""
import os
import stat
import tempfile

import pytest
from django.core.exceptions import ValidationError
from django.test import SimpleTestCase

from appsettings import (
    DictKeysTypeValidator,
    DictValuesTypeValidator,
    FileValidator,
    TypeValidator,
    ValuesTypeValidator,
)


class TypeValidatorTestCase(SimpleTestCase):
    """Test TypeValidator."""

    def test_valid(self):
        TypeValidator(int)(42)
        TypeValidator(float)(4.5)
        TypeValidator(list)([])

    def test_invalid(self):
        with self.assertRaisesMessage(ValidationError, "Value None is not of type int."):
            TypeValidator(int)(None)


class ValuesTypeValidatorTestCase(SimpleTestCase):
    """Test ValuesTypeValidator."""

    def test_valid(self):
        ValuesTypeValidator(int)([42, 14, 1676])
        ValuesTypeValidator(int)({42, 14, 1676})
        ValuesTypeValidator(int)((42, 14, 1676))

    def test_invalid(self):
        with self.assertRaisesMessage(ValidationError, "Element None is not of type int."):
            ValuesTypeValidator(int)([42, None, 1676])


class DictKeysTypeValidatorTestCase(SimpleTestCase):
    """Test DictKeysTypeValidator."""

    def test_valid(self):
        DictKeysTypeValidator(int)({42: "a", 1676: "b"})

    def test_invalid(self):
        with self.assertRaisesMessage(ValidationError, "The key None is not of type int."):
            DictKeysTypeValidator(int)({42: "a", None: "b"})


class DictValuesTypeValidatorTestCase(SimpleTestCase):
    """Test DictValuesTypeValidator."""

    def test_valid(self):
        DictValuesTypeValidator(int)({"a": 42, "b": 1676})

    def test_invalid(self):
        with self.assertRaisesMessage(ValidationError, "Item b's value None is not of type int."):
            DictValuesTypeValidator(int)({"a": 42, "b": None})


class FileValidatorTestCase(SimpleTestCase):
    """Test FileValidator."""

    def test_exists_success(self):
        with tempfile.NamedTemporaryFile() as fd:
            FileValidator(os.F_OK)(fd.name)

    def test_exists_error(self):
        fd = tempfile.NamedTemporaryFile()
        fd.close()
        with pytest.raises(ValidationError, match=r"Insufficient permissions for the file .+\."):
            FileValidator(os.F_OK)(fd.name)

    def test_error_message(self):
        fd = tempfile.NamedTemporaryFile()
        fd.close()
        with pytest.raises(ValidationError, match=r"My own message for .*!"):
            FileValidator(os.F_OK, "My own message for %(value)s!")(fd.name)

    def test_read_perm_success(self):
        with tempfile.NamedTemporaryFile() as fd:
            FileValidator(os.R_OK)(fd.name)

    def test_read_perm_error(self):
        with tempfile.NamedTemporaryFile() as fd:
            current_perms = stat.S_IMODE(os.lstat(fd.name).st_mode)
            os.chmod(fd.name, current_perms & ~stat.S_IRUSR)
            with pytest.raises(ValidationError, match=r"Insufficient permissions for the file .+\."):
                FileValidator(os.R_OK)(fd.name)

    def test_write_perm_success(self):
        with tempfile.NamedTemporaryFile() as fd:
            FileValidator(os.W_OK)(fd.name)

    def test_write_perm_error(self):
        with tempfile.NamedTemporaryFile() as fd:
            current_perms = stat.S_IMODE(os.lstat(fd.name).st_mode)
            os.chmod(fd.name, current_perms & ~stat.S_IWUSR)
            with pytest.raises(ValidationError, match=r"Insufficient permissions for the file .+\."):
                FileValidator(os.W_OK)(fd.name)

    def test_exec_perm_success(self):
        with tempfile.NamedTemporaryFile() as fd:
            current_perms = stat.S_IMODE(os.lstat(fd.name).st_mode)
            os.chmod(fd.name, current_perms | stat.S_IXUSR)
            FileValidator(os.X_OK)(fd.name)

    def test_exec_perm_error(self):
        with tempfile.NamedTemporaryFile() as fd:
            with pytest.raises(ValidationError, match=r"Insufficient permissions for the file .+\."):
                FileValidator(os.X_OK)(fd.name)

    def test_all_permissions(self):
        with tempfile.NamedTemporaryFile() as fd:
            current_perms = stat.S_IMODE(os.lstat(fd.name).st_mode)
            os.chmod(fd.name, current_perms | stat.S_IXUSR)
            FileValidator(os.R_OK | os.W_OK | os.X_OK)(fd.name)



"""Basic set of setting validators."""
import os

from django.core.exceptions import ValidationError


class TypeValidator(object):
    """Validator which checks type of the value."""

    message = "Value %(value)s is not of type %(type)s."

    def __init__(self, value_type, message=None):
        self.value_type = value_type
        if message:
            self.message = message

    def __call__(self, value):
        if not isinstance(value, self.value_type):
            params = {"value": value, "type": self.value_type.__name__}
            raise ValidationError(self.message, params=params)


class ValuesTypeValidator(object):
    """Validator which checks types of iterable values."""

    message = "Element %(value)s is not of type %(type)s."

    def __init__(self, value_type, message=None):
        self.value_type = value_type
        if message:
            self.message = message

    def __call__(self, value):
        for element in value:
            if not isinstance(element, self.value_type):
                params = {"value": element, "type": self.value_type.__name__}
                raise ValidationError(self.message, params=params)


class DictKeysTypeValidator(object):
    """Validator which checks types of dict keys."""

    message = "The key %(key)s is not of type %(type)s."

    def __init__(self, key_type, message=None):
        self.key_type = key_type
        if message:
            self.message = message

    def __call__(self, value):
        for key in value:
            if not isinstance(key, self.key_type):
                params = {"key": key, "type": self.key_type.__name__}
                raise ValidationError(self.message, params=params)


class DictValuesTypeValidator(object):
    """Validator which checks types of dict values."""

    message = "Item %(key)s's value %(value)s is not of type %(type)s."

    def __init__(self, value_type, message=None):
        self.value_type = value_type
        if message:
            self.message = message

    def __call__(self, value):
        for key, element in value.items():
            if not isinstance(element, self.value_type):
                params = {"key": key, "value": element, "type": self.value_type.__name__}
                raise ValidationError(self.message, params=params)


class FileValidator(object):
    """Validator which checks file existence and permissions."""

    message = "Insufficient permissions for the file %(value)s."

    def __init__(self, mode, message=None):
        """
        Validator initialization.

        Args:
            mode (optional int): Required permission for the file. Values might consist of inclusive OR between F_OK,
                R_OK, W_OK and X_OK constants. When checking read, write or execute permission, file existence (F_OK) is
                implicitly required. For more information check ``os.access`` documentation
                (https://docs.python.org/3/library/os.html#os.access).
            message (str): Override default ``ValidationError`` message.
        """
        self.mode = mode
        if message:
            self.message = message

    def __call__(self, value):
        """Validate the ``value``."""
        if not os.access(value, self.mode):
            raise ValidationError(self.message, params={"value": value})



"""Django AppSettings package."""
import os

from django.core.exceptions import ImproperlyConfigured
from django.core.signals import setting_changed

from .settings import (
    BooleanSetting,
    CallablePathSetting,
    DictSetting,
    FileSetting,
    FloatSetting,
    IntegerSetting,
    IterableSetting,
    ListSetting,
    NestedDictSetting,
    NestedListSetting,
    NestedSetting,
    ObjectSetting,
    PositiveFloatSetting,
    PositiveIntegerSetting,
    SetSetting,
    Setting,
    StringSetting,
    TupleSetting,
)
from .validators import (
    DictKeysTypeValidator,
    DictValuesTypeValidator,
    FileValidator,
    TypeValidator,
    ValuesTypeValidator,
)

__all__ = (
    "BooleanSetting",
    "CallablePathSetting",
    "DictKeysTypeValidator",
    "DictSetting",
    "DictValuesTypeValidator",
    "FileSetting",
    "FileValidator",
    "FloatSetting",
    "IntegerSetting",
    "IterableSetting",
    "ListSetting",
    "NestedDictSetting",
    "NestedListSetting",
    "NestedSetting",
    "ObjectSetting",
    "PositiveFloatSetting",
    "PositiveIntegerSetting",
    "SetSetting",
    "Setting",
    "StringSetting",
    "TupleSetting",
    "TypeValidator",
    "ValuesTypeValidator",
)


class _Metaclass(type):
    """
    ``AppSettings``'s metaclass.

    Each setting object declared in the class will be populated (name, prefix)
    and moved into the _meta.settings dictionary. A reference to this
    dictionary will also be added in the class as ``settings``.
    """

    def __new__(mcs, cls, bases, dct):
        """
        New method.

        Args:
            cls (str): class name.
            bases (tuple): base classes to inherit from.
            dct (dict): class attributes.

        Returns:
            class: the new created class.
        """
        super_new = super(_Metaclass, mcs).__new__

        # Also ensure initialization is only performed for subclasses
        # of AppSettings (excluding AppSettings class itself).
        parents = [b for b in bases if isinstance(b, _Metaclass)]
        if not parents:
            return super_new(mcs, cls, bases, dct)

        new_attr = {}
        _meta = dct.pop("Meta", type("Meta", (), {"setting_prefix": ""}))()
        _meta.settings = {}

        for name, setting in dct.items():
            if isinstance(setting, Setting):
                _meta.settings[name] = setting
                # populate name
                if setting.name == "":
                    setting.name = name
                # populate prefix
                if setting.prefix == "":
                    setting.prefix = _meta.setting_prefix
            else:
                new_attr[name] = setting
        new_attr["_meta"] = _meta
        new_attr["settings"] = _meta.settings

        return super_new(mcs, cls, bases, new_attr)

    def __getattr__(cls, item):
        """
        Return a setting object if it is in the ``_meta.settings`` dictionary.

        Args:
            item (str):
                the name of the setting variable (not the setting's name).

        Returns:
            ``Setting``: the setting object.

        Raises:
            AttributeError if the setting does not exist.
        """
        if item in cls._meta.settings.keys():
            return cls._meta.settings[item]
        raise AttributeError("'%s' class has no attribute '%s'" % (cls.__name__, item))


class AppSettings(metaclass=_Metaclass):
    """
    Base class for application settings.

    Only use this class as a parent class for inheritance. If you try to
    access settings directly in ``AppSettings``, it will raise a
    RecursionError. Some protections have been added to prevent you from
    instantiating this very class, or to return immediately when running
    ``AppSettings.check()``, but trying to access attributes on the class is
    not yet prevented.

    """

    OS_ENVIRON_OVERRIDE_PREFIX = "__DAP_"  # type: str

    def __init__(self):
        """
        Initialization method.

        The ``invalidate_cache`` and ``manage_environ_invalidation`` methods will be connected to the Django
        ``setting_changed`` signal in this method, with the dispatch UIDs being method initials and the id of this very
        object (``id(self)``).
        """
        if self.__class__ == AppSettings:
            raise RuntimeError("Do not use AppSettings class as itself, " "use it as a base for subclasses")
        setting_changed.connect(self.invalidate_cache, dispatch_uid="ic" + str(id(self)))
        setting_changed.connect(self.manage_environ_invalidation, dispatch_uid="mei" + str(id(self)))
        self._cache = {}

    def __getattr__(self, item):
        """
        Return a setting value.

        The caching is done here. If the setting exists, and if it's variable
        name is in the cache dictionary, return the cached value. If there
        is no cached value, get the setting value with ``setting.get_value()``,
        cache it, and return it.

        Args:
            item (str):
                the name of the setting variable (not the setting's name).

        Returns:
            object: a setting value.

        Raises:
            AttributeError if the setting does not exist.
        """
        if item in self.settings.keys():
            if item in self._cache:
                return self._cache[item]
            value = self._cache[item] = self.settings[item].get_value()
            return value
        raise AttributeError("'%s' object has no attribute '%s'" % (repr(self), item))

    @classmethod
    def check(cls):
        """
        Class method to check every settings.

        Will raise an ``ImproperlyConfigured`` exception with explanation.
        """
        if cls == AppSettings:
            return None

        exceptions = []
        for setting in cls.settings.values():
            try:
                setting.check()
            except ImproperlyConfigured as error:
                exceptions.append(str(error))
        if exceptions:
            raise ImproperlyConfigured("\n".join(exceptions))

    def manage_environ_invalidation(self, *, setting, enter, **kwargs):
        """
        Manage keys and values in ``os.environ`` on setting change.

        Environ values take precedence over ``django.conf`` values by default. But when we override a setting, we want
        to take the ``django.conf`` value and therefore we need to remove corresponding key from the ``os.environ``.
        To be able to restore the original value later, we add a prefix (OS_ENVIRON_OVERRIDE_PREFIX) to the key and then
        just remove the prefix.
        """
        for item in self.settings.keys():
            if self.settings[item].full_name == setting:
                setting_override_key = self.OS_ENVIRON_OVERRIDE_PREFIX + setting
                if enter and setting in os.environ:
                    os.environ[setting_override_key] = os.environ[setting]
                    del os.environ[setting]
                elif not enter and setting_override_key in os.environ:
                    os.environ[setting] = os.environ[setting_override_key]
                    del os.environ[setting_override_key]
                break

    def invalidate_cache(self, **kwargs):
        """Invalidate cache. Run when receive ``setting_changed`` signal."""
        self._cache = {}
